"""Xspec model primitives for JAX."""

from __future__ import annotations

from abc import ABC, abstractmethod
from functools import partial
from math import prod

import jax
import jax.numpy as jnp
from jax.core import Primitive, ShapedArray
from jax.interpreters import ad, batching, mlir, xla
from jax.lib import xla_client
from jaxlib.hlo_helpers import custom_call

import xspex


def avals_to_layouts(avals):
    return [list(reversed(range(aval.ndim))) for aval in avals]


class XspecPrimitiveBase(Primitive, ABC):
    def __init__(self, name: str):
        name = str(name)
        if name not in xspex.list_models():
            raise ValueError(f"Xspec v{xspex.version()} has no '{name}' model")

        super().__init__(f'XS{name}')
        self._model_name = name
        self._nparam = len(xspex.info(name).parameters)

        self.def_impl(partial(xla.apply_primitive, self))
        # self.def_impl(getattr(xspex, name))

        self.def_abstract_eval(self.get_abstract_eval())

        mlir.register_lowering(self, self.lowering, platform='cpu')

        batching.primitive_batchers[self] = self.batching

        ad.primitive_jvps[self] = self.jvp

    @abstractmethod
    def __call__(self, *args, **kwargs):
        pass

    @abstractmethod
    def get_abstract_eval(self):
        pass

    @abstractmethod
    def lowering(self, *args, **kwargs):
        pass

    @abstractmethod
    def batching(self, vector_arg_values, batch_axes):
        pass

    def jvp(self, primals, tangents):
        raise NotImplementedError(f'JVP is not implemented for {self}')


class XspecPrimitive(XspecPrimitiveBase):
    def __call__(self, params, egrid, spec_num):
        return self.bind(params, egrid, spec_num)

    def get_abstract_eval(self):
        name = self.name
        npar = self._nparam

        def abstract_eval(params, egrid, spec_num):
            params_shape = jnp.shape(params)
            if params_shape == () or params_shape[-1] != npar:
                raise ValueError(f'{name} params shape must be (..., {npar})')
            egrid_shape = jnp.shape(egrid)
            if len(egrid_shape) != 1:
                raise ValueError('egrid must be 1D array')
            spec_num_shape = jnp.shape(spec_num)
            if len(spec_num_shape):
                raise ValueError('spec_num must be a integer')

            shape = params_shape[:-1] + (egrid_shape[-1] - 1,)
            dtype = jax.dtypes.canonicalize_dtype(egrid.dtype)
            return ShapedArray(shape=shape, dtype=dtype)

        return abstract_eval

    def lowering(self, ctx, params, egrid, spec_num):
        egrid_type = mlir.ir.RankedTensorType(egrid.type)
        etype = egrid_type.element_type
        if isinstance(etype, mlir.ir.F32Type):
            call_target_name = f'{self._model_name}_f32'
        elif isinstance(etype, mlir.ir.F64Type):
            call_target_name = f'{self._model_name}_f64'
        else:
            raise NotImplementedError(f'unsupported dtype {etype}')
        out_shape = ctx.avals_out[0].shape
        out_type = mlir.ir.RankedTensorType.get(out_shape, etype)
        out_n = mlir.ir_constant(out_shape[-1])
        out_batch = mlir.ir_constant(prod(out_shape[:-1]))
        return custom_call(
            call_target_name,
            result_types=[out_type],
            operands=[params, egrid, spec_num, out_n, out_batch],
            operand_layouts=avals_to_layouts(ctx.avals_in) + [()] * 2,
            result_layouts=avals_to_layouts(ctx.avals_out)
        ).results

    def batching(self, vector_arg_values, batch_axes):
        if batch_axes[1] is not None:
            raise NotImplementedError('egrid batching is not implemented')
        if batch_axes[2] is not None:
            raise NotImplementedError('spec_num batching is not implemented')

        params, egrid, spec_num = vector_arg_values
        params = jnp.moveaxis(params, batch_axes[0], 0)
        return self(params, egrid, spec_num), 0

    def jvp(self, primals, tangents):
        params, egrid, spec_num = primals
        d_params = tangents[0]

        primals_out = self(params, egrid, spec_num)
        f_vmap = jax.vmap(jax.jit(self), in_axes=(0, None, None), out_axes=0)
        eps = jnp.finfo(params.dtype).eps
        identity = jnp.eye(len(params))
        params_abs = jnp.where(
            jnp.equal(params, 0.0),
            jnp.ones_like(params),
            jnp.abs(params)
        )
        params_abs = jnp.expand_dims(params_abs, axis=-1)

        # See Numerical Recipes Chapter 5.7
        USE_CENTRAL_DIFF = True
        if USE_CENTRAL_DIFF:
            delta = params_abs * eps ** (1.0 / 3.0)
            params_pos_perturb = params + identity * delta
            out_pos_perturb = f_vmap(params_pos_perturb, egrid, spec_num)
            params_neg_perturb = params - identity * delta
            out_neg_perturb = f_vmap(params_neg_perturb, egrid, spec_num)
            d_out = (out_pos_perturb - out_neg_perturb) / (2.0 * delta)
        else:
            delta = params_abs * jnp.sqrt(eps)
            params_perturb = params + identity * delta
            out_perturb = f_vmap(params_perturb, egrid, spec_num)
            d_out = (out_perturb - primals_out) / delta

        tangents_out = d_params @ d_out
        return primals_out, tangents_out


class XspecConvPrimitive(XspecPrimitiveBase):
    def __call__(self, params, egrid, flux, spec_num):
        return self.bind(params, egrid, flux, spec_num)

    def get_abstract_eval(self):
        name = self.name
        npar = self._nparam

        def abstract_eval(params, egrid, flux, spec_num):
            params_shape = jnp.shape(params)
            if params_shape == () or params_shape[-1] != npar:
                raise ValueError(f'{name} params shape must be (..., {npar})')
            egrid_shape = jnp.shape(egrid)
            if len(egrid_shape) != 1:
                raise ValueError('egrid must be 1D array')
            flux_shape = jnp.shape(flux)
            if flux_shape == () or flux_shape[-1] != egrid_shape[-1] - 1:
                raise ValueError('flux shape must be (..., len(egrid) - 1)')
            spec_num_shape = jnp.shape(spec_num)
            if len(spec_num_shape):
                raise ValueError('spec_num must be a integer')

            if len(params_shape) > 1 and len(flux_shape) > 1:
                if params_shape[:-1] != flux_shape[:-1]:
                    raise ValueError(
                        'params and flux should have the same batch size'
                    )
            if len(params_shape) >= len(flux_shape):
                shape = params_shape[:-1] + (egrid_shape[-1] - 1,)
            else:
                shape = flux_shape[:-1] + (egrid_shape[-1] - 1,)
            dtype = jax.dtypes.canonicalize_dtype(egrid.dtype)
            return ShapedArray(shape=shape, dtype=dtype)

        return abstract_eval

    def lowering(self, ctx, params, egrid, flux, spec_num):
        egrid_type = mlir.ir.RankedTensorType(egrid.type)
        etype = egrid_type.element_type
        if isinstance(etype, mlir.ir.F32Type):
            call_target_name = f'{self._model_name}_f32'
        elif isinstance(etype, mlir.ir.F64Type):
            call_target_name = f'{self._model_name}_f64'
        else:
            raise NotImplementedError(f'unsupported dtype {etype}')

        params_shape = ctx.avals_in[0].shape
        params_batch = mlir.ir_constant(prod(params_shape[:-1]))

        flux_shape = ctx.avals_in[2].shape
        flux_batch = mlir.ir_constant(prod(flux_shape[:-1]))

        out_shape = ctx.avals_out[0].shape
        out_type = mlir.ir.RankedTensorType.get(out_shape, etype)
        out_n = mlir.ir_constant(out_shape[-1])

        return custom_call(
            call_target_name,
            result_types=[out_type],
            operands=[
                params, egrid, flux, spec_num, out_n, params_batch, flux_batch
            ],
            operand_layouts=avals_to_layouts(ctx.avals_in) + [()] * 3,
            result_layouts=avals_to_layouts(ctx.avals_out)
        ).results

    def batching(self, vector_arg_values, batch_axes):
        if batch_axes[1] is not None:
            raise NotImplementedError('egrid batching is not implemented')
        if batch_axes[3] is not None:
            raise NotImplementedError('spec_num batching is not implemented')

        params, egrid, flux, spec_num = vector_arg_values
        if batch_axes[0] is not None:
            params = jnp.moveaxis(params, batch_axes[0], 0)
        if batch_axes[2] is not None:
            flux = jnp.moveaxis(flux, batch_axes[2], 0)
        return self(params, egrid, flux, spec_num), 0


def get_primitive(
    model: str
) -> tuple[XspecPrimitive | XspecConvPrimitive, xspex.XspecModel]:
    """Return the primitive for the given Xspec model.

    Parameters
    ----------
    model : str
        The Xspec model name.

    Returns
    -------
    primitive : XspecPrimitive or XspecConvPrimitive
        The primitive for the model.
    model : XspecModel
        The dataclass that describes the model.

    Examples
    --------

    >>> apec, info = get_primitive('apec')
    >>> apec
    'apec'

    """
    model = str(model)
    check = model.casefold()
    p = next(
        (v for k, v in XSModel['primitive'].items() if k.casefold() == check),
        None
    )
    if p is None:
        raise ValueError(f"Unrecognized Xspec model '{model}'")

    return p, xspex.info(model)


for k, v in xspex.xla_registrations().items():
    xla_client.register_custom_call_target(k, v, platform='cpu')

add = xspex.list_models(xspex.ModelType.Add)
mul = xspex.list_models(xspex.ModelType.Mul)
con = xspex.list_models(xspex.ModelType.Con)
primitive = {m: XspecPrimitive(m) for m in add + mul}
primitive |= {m: XspecConvPrimitive(m) for m in con}
XSModel = {
    'add': {m: xspex.info(m) for m in add},
    'mul': {m: xspex.info(m) for m in mul},
    'con': {m: xspex.info(m) for m in con},
    'primitive': primitive,
}
del k, v, add, mul, con, primitive



import os
import pathlib
import re
import subprocess


def get_compiler():
    compiler = os.getenv("CXX")
    if compiler is not None:
        return compiler

    # We could try and be clever here, but for now just fall through
    # to g++ and assume that anyone using clang has to set CXX.  I do
    # not build on macOS so it's not a problem for me^(TM).
    #
    compiler = "g++"
    return compiler


def compile_code(HEADAS):
    """Compile the code."""

    basename = "report_xspec_version"
    helpers = pathlib.Path("helpers")

    compiler = get_compiler()
    args = [compiler, str(helpers / f"{basename}.cxx"),
            "-o", str(helpers / basename),
            f"-I{HEADAS}/include",
            f"-L{HEADAS}/lib", "-lXSUtil"
        ]

    subprocess.run(args, check=True)
    return helpers / basename


def get_xspec_macros(HEADAS):
    """Return the macro definitions which define the Xspec version.

    Parameters
    ----------
    HEADAS : pathlib.Path
        The path to the HEADAS location.

    Returns
    -------
    xspec_version, macros : str, list
        The Xspec version, including the patch level, and then the
        macro definitions to pass to the compiled code.

    """

    code = compile_code(HEADAS)
    command = subprocess.run([str(code)],
                             check=True,
                             stdout=subprocess.PIPE)

    xspec_version = command.stdout.decode().strip()
    print(f"Building against Xspec: '{xspec_version}'")

    # split the Xspec version
    toks = xspec_version.split(".")
    assert len(toks) == 3, xspec_version
    xspec_major = toks[0]
    xspec_minor = toks[1]

    match = re.match(r"^(\d+)(.*)$", toks[2])
    xspec_micro = match[1]
    xspec_patch = None if match[2] == "" else match[2]

    macros = [
        ('BUILD_Xspec', xspec_version),
        ('BUILD_Xspec_MAJOR', xspec_major),
        ('BUILD_Xspec_MINOR', xspec_minor),
        ('BUILD_Xspec_MICRO', xspec_micro),
    ]

    if xspec_patch is not None:
        macros.append(('BUILD_Xspec_PATCH', xspec_patch))

    return xspec_version, macros



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto
import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence

    from .primitive import XspecPrimitive, XspecConvPrimitive

try:
    from ._compiled import *
    __version__ = _compiled.__version__
    __INITIALIZED__ = True
except ImportError as ie:
    # Allow the actual error message to be reported if the user
    # has tweaked the log level, for instance with:
    #
    #   import logging; logging.basicConfig(level=logging.DEBUG)
    #
    logging.getLogger(__name__).warn("Unable to import compiled Xspec models")
    logging.getLogger(__name__).info(str(ie))

    __version__ = "none"
    __INITIALIZED__ = False

_compiled._init()


class ModelType(Enum):
    """The various Xspec model types."""

    Add = auto()
    Mul = auto()
    Con = auto()


class LanguageStyle(Enum):
    """The various ways to define and call Xspec models."""

    CppStyle8 = auto()
    CStyle8 = auto()
    F77Style4 = auto()
    F77Style8 = auto()


class ParamType(Enum):
    """The Xspec parameter type."""

    Default = auto()
    Switch = auto()
    Scale = auto()
    Periodic = auto()


@dataclass
class XspecParameter:
    """An Xspec parameter."""

    paramtype: ParamType
    name: str
    default: float
    units: str | None = None
    frozen: bool = False
    # Would it be better to just have limits = [hardmin, softmni, softmax, hardmax]?
    softmin: float | None = None
    softmax: float | None = None
    hardmin: float | None = None
    hardmax: float | None = None
    delta: float | None = None


@dataclass
class XspecModel:
    """An Xspec model."""

    modeltype: ModelType
    name: str
    funcname: str
    language: LanguageStyle
    elo: float
    ehi: float
    parameters: Sequence[XspecParameter]
    use_errors: bool = False
    can_cache: bool = True


_info = {
##PYINFO##
}


def info(model: str) -> XspecModel:
    """Return information on the Xspec model from the model.dat file.

    This returns the information on the model as taken from the Xspec
    model library used to build this model.

    Parameters
    ----------
    name : str
        The Xspec model name (case-insensitive).

    Returns
    -------
    model : XspecModel
        The dataclass that describes the mode;.

    See Also
    --------
    list_models

    Examples
    --------

    >>> m = info('apec')
    >>> m.name
    'apec'
    >>> m.modeltype
    <ModelType.Add: 1>
    >>> [(p.name, p.default, p.units) for p in m.parameters]
    [('kT', 1.0, 'keV'), ('Abundanc', 1.0, None), ('Redshift', 0.0, None)]

    """
    # We want case-insensitive comparison but for the keys to retain
    # their case. Using casefold() rather than lower() is a bit OTT
    # here as I would bet model.dat is meant to be US-ASCII.
    model = str(model)
    check = model.casefold()
    out = next((v for k, v in _info.items() if k.casefold() == check), None)
    if out is None:
        raise ValueError(f"Unrecognized Xspec model '{model}'")

    return out


# Do we need Optional here?
def list_models(modeltype: ModelType | None = None,
                language: LanguageStyle | None = None) -> list[str]:
    """Returns the names of Xspec models from the model.dat file.

    This returns the information on the model as taken from the Xspec
    model library used to build this model.

    Parameters
    ----------
    modeltype : ModelType or None, optional
        If not None then restrict the list to this model type.
    language : LanguageStyle or None, optional
        If not None then restrict the list to this language type.

    Returns
    -------
    models : list of str
        A sorted list of the model names.

    See Also
    --------
    info

    Notes
    -----
    The restrictions are combined, so setting both `modeltype` and
    `language` will select only those models which match both filters.

    Examples
    --------

    >>> len(list_models())
    231

    >>> 'tbabs' in list_models()
    False
    >>> 'TBabs' in list_models()
    True

    With Xspec 12.13.1 / HEASOFT 6.29:

    >>> list_models(modeltype=ModelType.Con)
    ['cflux', 'clumin', 'cpflux', 'gsmooth', ..., 'zashift', 'zmshift']

    >>> list_models(modeltype=ModelType.Con, language=LanguageStyle.F77Style4)
    ['kyconv', 'rgsxsrc', 'thcomp']

    """

    out = set()
    for k, v in _info.items():

        if modeltype is not None and v.modeltype != modeltype:
            continue

        if language is not None and v.language != language:
            continue

        out.add(k)

    return sorted(out)


if __INITIALIZED__:
    from .primitive import get_primitive



"""Generate the code to be built based on the model.dat file.

Note that any error in creating the template will raise a
SystemExit exception.

"""

import pathlib
import sys

from parse_xspec.models import parse_xspec_model_description


__all__ = ("apply", )


def replace_term(txt, term, replace):
    """Replace the first ``term`` with ``replace`` in txt"""

    idx = txt.find(term)
    if idx < 0:
        sys.stderr.write(f'ERROR: unable to find {term}\n')
        sys.exit(1)

    return txt[:idx] + replace + txt[idx + len(term):]


def apply_compiled(models, template, outfile):
    """Convert the template for the compiled code."""

    # Ugly information flow here
    addmodels = []
    mulmodels = []
    conmodels = []
    f77models = []
    f77models_dp = []
    cxxmodels = []
    cmodels = []

    def get_npars(npars):
        if npars == 0:
            return "no parameters"
        if npars == 1:
            return "1 parameter"

        return f"{npars} parameters"

    def wrapmodel_basic(model, npars, call, text,
                        inplace=False,
                        convolution=False):
        """Create the m.def line for a single model"""

        assert not(inplace and convolution)

        out = f'    m.def("{model.name}", {call}'

        if model.language == 'Fortran - single precision':
            out += f'_f<float, {model.funcname}_'
            f77models.append(model.funcname)
        elif model.language == 'Fortran - double precision':
            out += F'_F<double, {model.funcname}_'
            f77models_dp.append(model.funcname)
        elif model.language == 'C++ style':
            out += f'_C<double, C_{model.funcname}'
            cxxmodels.append(model.funcname)
        elif model.language == 'C style':
            out += f'_C<double, {model.funcname}'  # should this be 'c_{model.funcname}' (not for compmag...)?
            cmodels.append(model.funcname)
        else:
            assert False, (model.name, model.funcname, model.language)

        out += f', {npars}>, "{text}",'
        out += '"pars"_a, "energies"_a, '
        if convolution:
            out += '"model"_a, '

        if inplace:
            out += '"out"_a, '

        out += '"spectrum"_a=1'
        if not model.language.startswith('Fortran'):
            out += ', "initStr"_a=""'

        # if inplace or convolution:
        #     out += ',py::return_value_policy::reference'

        out += ');'
        return out

    # def wrapmodel_cxx(model, npars, text):
    #     """Make the C++ version available as name_"""
    #
    #     if model.language != 'C++ style':
    #         return ''
    #
    #     out = '\n'
    #     out += f'    m.def("{model.name}_", xspex::wrapper_inplace_CXX<'
    #     out += f'{model.funcname}, {npars}>, "{text}",'
    #     out += '"pars"_a,"energies"_a,"out"_a,"spectrum"_a=1,'
    #     out += '"initStr"_a=""'
    #     out += ',py::return_value_policy::reference'
    #     out += ');'
    #     return out

    def wrapmodel_add(model, npars):
        """What is the m.def line for this additive model?"""

        addmodels.append((model.name, npars))

        out = wrapmodel_basic(model, npars, 'xspex::wrapper',
                              f'The Xspec additive {model.name} model with {get_npars(npars)}.')
        # out += '\n'
        # out += wrapmodel_basic(model, npars, 'xspec_models_cxc::wrapper_inplace',
        #                        f'The Xspec additive {model.name} model ({get_npars(npars)}); inplace.',
        #                        inplace=True)

        # out += wrapmodel_cxx(model, npars,
        #                      f"The Xspec additive {model.name} model ({get_npars(npars)}); RealArray, inplace.")

        return out

    def wrapmodel_mul(model, npars):
        """What is the m.def line for this multiplicative model?"""

        mulmodels.append((model.name, npars))

        out = wrapmodel_basic(model, npars, 'xspex::wrapper',
                              f'The Xspec multiplicative {model.name} model with {get_npars(npars)}.')
        # out += '\n'
        # out += wrapmodel_basic(model, npars, 'xspec_models_cxc::wrapper_inplace',
        #                        f'The Xspec multiplicative {model.name} model ({get_npars(npars)}); inplace.',
        #                        inplace=True)
        #
        # out += wrapmodel_cxx(model, npars,
        #                      f"The Xspec multiplicative {model.name} model ({get_npars(npars)}); RealArray, inplace.")

        return out

    def wrapmodel_con(model, npars):
        """What is the m.def line for this convolution model?"""

        conmodels.append((model.name, npars))

        out = wrapmodel_basic(model, npars, 'xspex::wrapper_con',
                              f'The Xspec convolution {model.name} model with {get_npars(npars)}.',
                              convolution=True)

        return out

    def wrapmodel(model):
        """What is the m.def line for this model?"""

        npars = len(model.pars)
        if model.modeltype == 'Add':
            return wrapmodel_add(model, npars)

        if model.modeltype == 'Mul':
            return wrapmodel_mul(model, npars)

        if model.modeltype == 'Con':
            return wrapmodel_con(model, npars)

        assert False, model.modeltype

    mstrs = [wrapmodel(m) for m in models]

    def wrapxla_basic(model, npars, call, dtype):
        """Create the xla line for a single model"""
        assert dtype in ['float', 'double']
        if dtype == 'float':
            type_str = 'f32'
        else:
            type_str = 'f64'

        out = f'    reg["{model.name}_{type_str}"] = {call}'

        if model.modeltype == 'Con':
            out += '_con'

        if model.language == 'Fortran - single precision':
            out += f'_f_XLA_{type_str}<{model.funcname}_'
            f77models.append(model.funcname)
        elif model.language == 'Fortran - double precision':
            out += F'_F_XLA_{type_str}<{model.funcname}_'
            f77models_dp.append(model.funcname)
        elif model.language == 'C++ style':
            out += f'_C_XLA_{type_str}<C_{model.funcname}'
            cxxmodels.append(model.funcname)
        elif model.language == 'C style':
            out += f'_C_XLA_{type_str}<{model.funcname}'
            cmodels.append(model.funcname)
        else:
            assert False, (model.name, model.funcname, model.language)

        out += f', {npars}>);'
        return out

    def wrapxla(model, dtype):
        return wrapxla_basic(model, len(model.pars), 'xspex::EncapsulateFunction(xspex::wrapper', dtype)

    xla_float_regs = [wrapxla(m, 'float') for m in models]
    xla_double_regs = [wrapxla(m, 'double') for m in models]

    def hdr(arg):
        (model, npars) = arg
        return f"{model} - {get_npars(npars)}."

    with template.open(mode='rt') as ifh:
        out = ifh.read()

        repl = [hdr(m) for m in addmodels]
        out = replace_term(out, '/*ADDMODELS*/', '\n'.join(repl))

        repl = [hdr(m) for m in mulmodels]
        out = replace_term(out, '/*MULMODELS*/', '\n'.join(repl))

        repl = [hdr(m) for m in conmodels]
        out = replace_term(out, '/*CONMODELS*/', '\n'.join(repl))

        out = replace_term(out, '/*MODELS*/', '\n'.join(mstrs))

        out = replace_term(out, '/*XLA_FLOAT_MODELS*/', '\n'.join(xla_float_regs))
        out = replace_term(out, '/*XLA_DOUBLE_MODELS*/', '\n'.join(xla_double_regs))

    with outfile.open(mode='wt') as ofh:
        ofh.write(out)

    return {'outfile': outfile,
            'models': [m.name for m in models],
            'additive': addmodels,
            'multiplicative': mulmodels,
            'convolution': conmodels,
            'C++': cxxmodels,
            'C': cmodels,
            'f77': f77models,
            'F77': f77models_dp,
        }


def apply_python(models, template, xspec_version, outfile):
    """Convert the template for the Python code.

    xspec_version : str
        The Xspec library we are building against, including the patch
        level.

    """

    def to_model(mtype):
        return {'Add': 'ModelType.Add',
                'Mul': 'ModelType.Mul',
                'Con': 'ModelType.Con' }[mtype]

    def to_lang(langtype):
        return {'C++ style': 'LanguageStyle.CppStyle8',
                'C style': 'LanguageStyle.CStyle8',
                'Fortran - single precision': 'LanguageStyle.F77Style4',
                'Fortran - double precision': 'LanguageStyle.F77Style8'}[langtype]

    def to_ptype(ptype):
        # We don't support periodic yet
        return {'Basic': 'ParamType.Default',
                'Switch': 'ParamType.Switch',
                'Scale': 'ParamType.Scale',
                '?': 'ParamType.Periodic'}[ptype]

    mstrs = []
    for model in models:
        mtype = to_model(model.modeltype)
        lang = to_lang(model.language)
        out = [f"    '{model.name}': XspecModel(modeltype={mtype}",
               f"name='{model.name}'",
               f"funcname='{model.funcname}'",
               f"language={lang}",
               f"elo={model.elo}",
               f"ehi={model.ehi}"]

        pars = []
        for p in model.pars:
            ps = [f"XspecParameter(paramtype={to_ptype(p.paramtype)}"]
            ps.append(f"name='{p.name}'")
            ps.append(f"default={p.default}")
            if p.units is not None:
                ps.append(f"units='{p.units}'")

            try:
                if p.frozen:
                    ps.append("frozen=True")
            except AttributeError:
                # Assume that if there's no frozen attribute, it is
                # always frozen
                ps.append("frozen=True")

            for t in ['soft', 'hard']:
                for r in ['min', 'max']:
                    attr = getattr(p, f'{t}{r}')
                    if attr is not None:
                        ps.append(f"{t}{r}={attr}")

            if p.delta is not None:
                ps.append(f"delta={p.delta}")

            pars.append(', '.join(ps) + ')')

        pars = ', '.join(pars)
        out.append(f"parameters=[{pars}]")

        if len(model.flags) > 0 and model.flags[0] > 0:
            out.append("use_errors=True")

        if len(model.flags) > 1 and model.flags[1] > 0:
            out.append("can_cache=False")

        mstrs.append(', '.join(out) + ')')

    with template.open(mode='rt') as ifh:
        out = ifh.read()

        out = replace_term(out, '##PYINFO##', ',\n'.join(mstrs))
        # out = replace_term(out, '@@XspecVER@@', xspec_version)

    with outfile.open(mode='wt') as ofh:
        ofh.write(out)


# At what point does it become worth using a templating system like
# Jinja2?
#
def apply(modelfile, xspec_version, out_dir):
    """Parse the model.dat file and create the code.

    Parameters
    ----------
    modelfile : pathlib.Path
        The model.dat file.
    xspec_version : str
        The Xspec library we are building against, including the patch
        level.
    out_dir : pathlib.Path
        The directory where to write the output.

    Returns
    -------
    models : dict
        Information on the different model types that were found and
        the files created.

    Notes
    -----
    The templates/xspec.cxx file is used as the template.

    """

    if not modelfile.is_file():
        sys.stderr.write(f'ERROR: unable to find model.dat file: {modelfile}.\n')
        sys.exit(1)

    allmodels = parse_xspec_model_description(modelfile)
    if len(allmodels) == 0:
        sys.stderr.write(f'ERROR: unable to parse model.fat file: {modelfile}\n')
        sys.exit(1)

    # Filter to the ones we care about
    models = [m for m in allmodels if m.modeltype in ['Add', 'Mul', 'Con']]

    # For now filter out FORTRAN double-precision models as I don't
    # know what to do with them (need to check against Sherpa).
    #
    allowed_langs = ["Fortran - single precision",
                     "Fortran - double precision",
                     "C style",
                     "C++ style"]
    models = [m for m in models if m.language in allowed_langs]

    if len(models) == 0:
        sys.stderr.write(f'ERROR: unable to find any models in: {modelfile}\n')
        sys.exit(1)

    if not out_dir.is_dir():
        sys.stderr.write(f'ERROR: unable to find output directory: {out_dir}\n')
        sys.exit(1)

    # Create the code we want to compile
    #
    template_dir = pathlib.Path('template')

    filename = pathlib.Path('xspex.cpp')
    template = template_dir / filename
    outfile = out_dir / filename
    out_dir.mkdir(exist_ok=True)
    info = apply_compiled(models, template, outfile)

    filename = pathlib.Path('__init__.py')
    template = template_dir / filename
    out_dir = out_dir / 'xspex'
    outfile = out_dir / filename
    out_dir.mkdir(exist_ok=True)
    apply_python(models, template, xspec_version, outfile)

    # It looks like info['models'] does not contain repeated values.
    print("###############################################")
    print(f"Number of supported models:  {len(info['models'])}")

    # Summarize. We can't use the length of these as they may include
    # repeated models (e.g. because we provide different ways to
    # access the model). Hence, the conversion to a set first (as the
    # names are the same).
    #
    def count(label, value=None):
        if value is None:
            value = label

        print(f"   {label:20s}  {len(set(info[value]))}")

    count("additive")
    count("multiplicative")
    count("convolution")
    count("C++")
    count("C")
    count("FORTRAN (sp)", "f77")
    count("FORTRAN (dp)", "F77")

    nskip = len(allmodels) - len(info['models'])
    if nskip > 0:
        print("")
        print(f"   Number skipped: {nskip}")

        assert len(info['models']) == len(models)

        allknown = set(m.name for m in allmodels)
        known = set(m.name for m in models)
        for i, unknown in enumerate(sorted(allknown.difference(known)), 1):
            print(f"      {i}. {unknown}")

        print("")

    print("###############################################")

    return info



import numpy as np
import pytest

import xspex as x

# We need to set up the cosmology as this is currently not done in FNINIT.
#
x.cosmology(H0=70, lambda0=0.73, q0=0)

# We want to ensure we have a fixed abundance / cross section
# for the checks below. If we didn't set them here then it
# would depend on the user's ~/.xspec/Xspec.init file
#
x.abundance('lodd')
x.cross_section('vern')


def get_dtype(model):
    """What is the drtype used by this model?"""

    return {x.LanguageStyle.CppStyle8: np.float64,
            x.LanguageStyle.CStyle8: np.float64,
            x.LanguageStyle.F77Style4: np.float32,
            x.LanguageStyle.F77Style8: np.float64}[model.language]


def test_have_version():
    """Minimal chekc of get_versino"""
    v = x.version()
    assert v != ''
    assert len(v.split('.')) == 3, v


def test_default_abundance():
    assert x.abundance() == 'lodd'


def test_default_cross_section():
    assert x.cross_section() == 'vern'


def test_number_elements():
    """Technically this could change in an XSPEC release"""
    assert x.number_elements == 30


@pytest.mark.parametrize("z,name",
                         [(1, 'H'), (2, 'He'), (17, 'Cl'), (29, 'Cu'), (30, 'Zn')])
def test_elementName(z, name):
    assert x.element_name(z) == name


@pytest.mark.parametrize("z,value",
                         [(1, 1.0),
                          (2, 0.07919999957084656),
                          (17, 1.8199999374246545e-07),
                          (29, 1.8199999729517913e-08),
                          (30, 4.2700001756657e-08)])
def test_elementAbundance(z, value):
    x.abundance('lodd')
    assert x.element_abundance(z) == pytest.approx(value)
    assert x.element_abundance(x.element_name(z)) == pytest.approx(value)


def test_cosmo_get():
    """Check the values we set above"""
    ans = x.cosmology()
    assert ans == pytest.approx({'H0': 70, 'lambda0': 0.73, 'q0': 0.0})


def test_has_wabs_info():
    """Check wabs.

    This is used as the model.dat entry is not likely to change
    for it.
    """

    assert 'wabs' in x.list_models()
    assert 'wabs' in x.list_models(modeltype=x.ModelType.Mul)
    assert 'wabs' not in x.list_models(modeltype=x.ModelType.Add)
    assert 'wabs' not in x.list_models(modeltype=x.ModelType.Con)

    model = x.info('wabs')
    assert model.name == 'wabs'
    assert model.modeltype == x.ModelType.Mul
    assert len(model.parameters) == 1

    # Technically we should be able to support this if it changes,
    # but it indicates that test_eval_wabs and test_eval_wabs_inline
    # may need the reference array changed.
    #
    assert model.language == x.LanguageStyle.F77Style4

    par = model.parameters[0]
    assert par.name == 'nH'
    assert par.default == pytest.approx(1.0)
    assert not par.frozen


WABS_MODEL = [8.8266723e-05, 2.3582002e-02, 1.5197776e-01]

@pytest.mark.parametrize("pars", [[0.1], (0.1, ), np.asarray([0.1])])
@pytest.mark.parametrize("energies", [[0.1, 0.2, 0.3, 0.4],
                                      (0.1, 0.2, 0.3, 0.4),
                                      np.arange(0.1, 0.5, 0.1)])
def test_eval_wabs(pars, energies):
    """Explicit tests of a model.

    This checks a "random" model - chosen as wabs as it's assumed it's
    not going to change - evaluates as expected.

    """

    y = x.wabs(energies=energies, pars=pars)

    # Let's assume this isn't going to change much
    #
    assert y == pytest.approx(WABS_MODEL)


# Unfortunately some models need to be skipped for
# some reason. This is obviously going to be version-specific.
#
# grbjet occasional failures has been reported to XSPEC (in 12.12.0).
#
# rfxconv and xilconv require additinal setup (e.g. energy range) that I have no
# energy to diesntangle, so we skip
#
# rgsxsrc requires extra setup (an image file)
#
MODELS_ADD = x.list_models(modeltype=x.ModelType.Add)
MODELS_MUL = x.list_models(modeltype=x.ModelType.Mul)
MODELS_CON = x.list_models(modeltype=x.ModelType.Con)

MODELS_ADD_SKIP = ['grbjet']
MODELS_MUL_SKIP = []
MODELS_CON_SKIP = ['rfxconv', 'rgsxsrc', 'xilconv']


@pytest.mark.parametrize("model", MODELS_ADD)
def test_eval_add(model):
    """Evaluate an additive model.

    Check we get a value > 0. To do so we ensure redshift > 0
    as this is an issue for some models.

    We skip models with can_cache set to False because we
    assume that the necessary setup has not been called
    """

    info = x.info(model)
    if not info.can_cache:
        pytest.skip(f"Model {model} can not be cached.")

    if model in MODELS_ADD_SKIP:
        pytest.skip(f"Model {model} is marked as un-testable.")

    mfunc = getattr(x, model)

    pars = [0.1 if p.name.casefold() == 'redshift' else p.default
            for p in info.parameters]

    egrid = np.arange(0.1, 10, 0.01)
    y1 = mfunc(energies=egrid, pars=pars)

    assert (y1 > 0).any()


@pytest.mark.parametrize("model", MODELS_MUL)
def test_eval_mul(model):
    """Evaluate a multiplicative model.
    """

    info = x.info(model)

    if model in MODELS_MUL_SKIP:
        pytest.skip(f"Model {model} is marked as un-testable.")

    mfunc = getattr(x, model)

    pars = [p.default for p in info.parameters]

    egrid = np.arange(0.1, 10, 0.01)
    y1 = mfunc(energies=egrid, pars=pars)

    assert (y1 > 0).any()


@pytest.mark.parametrize("model", MODELS_CON)
def test_eval_con(model):
    """Evaluate a convolution additive model."""

    info = x.info(model)
    if not info.can_cache:
        pytest.skip(f"Model {model} can not be cached.")

    if model in MODELS_CON_SKIP:
        pytest.skip(f"Model {model} is marked as un-testable.")

    egrid = np.arange(0.1, 10, 0.01)

    def conv(p):
        """what is the default parameter value?"""

        if p.name.casefold() == 'redshift':
            return 0.01

        if p.name.casefold() == 'velocity':
            return 100

        return p.default

    pars = [p.default for p in x.info('powerlaw').parameters]
    mvals = x.powerlaw(energies=egrid, pars=pars)
    mvals = mvals.astype(get_dtype(info))
    ymodel = mvals.copy()

    mfunc = getattr(x, model)

    pars = [conv(p) for p in info.parameters]
    y1 = mfunc(energies=egrid, pars=pars, model=mvals)

    assert (y1 > 0).any()
    assert (y1 != ymodel).any()



import jax
import jax.numpy as jnp
import numpy as np
import pytest

import xspex as x

jax.config.update('jax_enable_x64', True)

# We need to set up the cosmology as this is currently not done in FNINIT.
#
x.cosmology(H0=70, lambda0=0.73, q0=0)

# We want to ensure we have a fixed abundance / cross section
# for the checks below. If we didn't set them here then it
# would depend on the user's ~/.xspec/Xspec.init file
#
x.abundance('lodd')
x.cross_section('vern')


def test_powerlaw_primitive():
    @jax.jit
    def pl(p, e, _):
        p = 1 - p
        f = e ** p / p
        return f[1:] - f[:-1]

    powerlaw, _ = x.get_primitive('powerlaw')
    params = jnp.array([1.5])
    egrid = jnp.arange(1., 10., 0.1)

    res_true = pl(params, egrid, 1)
    res_test1 = powerlaw(params, egrid, 1)
    res_test2 = jax.jit(powerlaw)(params, egrid, 1)

    assert jnp.allclose(res_true, res_test1)
    assert jnp.allclose(res_true, res_test2)

    prim_in = (jnp.r_[1.1], jnp.linspace(1.1, 1.5, 6), 1)
    tan_in = (
        jnp.r_[1.0],
        jnp.zeros_like(prim_in[1]),
        np.zeros((), dtype=jax.dtypes.float0)
    )
    jvp_true = jax.jvp(pl, prim_in, tan_in)
    jvp_test = jax.jvp(powerlaw, prim_in, tan_in)
    assert jnp.allclose(jvp_true[0], jvp_test[0])
    assert jnp.allclose(jvp_true[1], jvp_test[1])

    args = (
        jnp.expand_dims(jnp.linspace(-3, 3, 101), axis=-1),
        jnp.geomspace(0.1, 20, 10001),
        1,
    )
    f1 = jax.vmap(jax.jit(powerlaw), (0, None, None), 0)
    f2 = jax.vmap(pl, (0, None, None), 0)
    assert jnp.allclose(f1(*args), f2(*args))


# Unfortunately some models need to be skipped for
# some reason. This is obviously going to be version-specific.
#
# grbjet occasional failures has been reported to XSPEC (in 12.12.0).
#
# rfxconv and xilconv require additinal setup (e.g. energy range) that I have no
# energy to diesntangle, so we skip
#
# rgsxsrc requires extra setup (an image file)
#
MODELS_ADD = x.list_models(modeltype=x.ModelType.Add)
MODELS_MUL = x.list_models(modeltype=x.ModelType.Mul)
MODELS_CON = x.list_models(modeltype=x.ModelType.Con)

MODELS_ADD_SKIP = ['grbjet']
MODELS_MUL_SKIP = []
MODELS_CON_SKIP = ['rfxconv', 'rgsxsrc', 'xilconv']


@pytest.mark.parametrize("model", MODELS_ADD + MODELS_MUL)
def test_eval(model):
    p, info = x.get_primitive(model)
    if not info.can_cache:
        pytest.skip(f"Model {model} can not be cached.")

    if model in MODELS_ADD_SKIP:
        pytest.skip(f"Model {model} is marked as un-testable.")

    mfunc = getattr(x, model)

    pars = [0.1 if p.name.casefold() == 'redshift' else p.default
            for p in info.parameters]
    pars = np.array(pars)

    egrid = np.arange(0.1, 10, 0.01)
    y1 = mfunc(energies=egrid, pars=pars)
    y2 = p(pars, egrid, 1)
    y3 = jax.jit(p)(pars, egrid, 1)

    assert np.allclose(y1, y2)
    assert np.allclose(y1, y3)


@pytest.mark.parametrize("model", MODELS_CON)
def test_eval_con(model):
    """Evaluate a convolution additive model."""
    p, info = x.get_primitive(model)
    if not info.can_cache:
        pytest.skip(f"Model {model} can not be cached.")

    if model in MODELS_CON_SKIP:
        pytest.skip(f"Model {model} is marked as un-testable.")

    egrid = np.arange(0.1, 10, 0.01)

    def conv(p):
        """what is the default parameter value?"""

        if p.name.casefold() == 'redshift':
            return 0.01

        if p.name.casefold() == 'velocity':
            return 100.0

        return p.default

    pars = [p.default for p in x.info('powerlaw').parameters]
    pars = np.array(pars)
    mvals = x.powerlaw(energies=egrid, pars=pars)
    # mvals = mvals.astype(get_dtype(info))
    ymodel = mvals.copy()

    mfunc = getattr(x, model)

    pars = np.array([conv(p) for p in info.parameters])
    y1 = mfunc(energies=egrid, pars=pars, model=mvals)
    y2 = p(pars, egrid, mvals, 1)
    y3 = jax.jit(p)(pars, egrid, mvals, 1)

    assert (y1 > 0).any()
    assert np.any(y1 != ymodel)
    assert np.allclose(y1, y2)
    assert np.allclose(y1, y3)


@pytest.mark.parametrize("model", MODELS_ADD + MODELS_MUL)
def test_batching(model):
    p, info = x.get_primitive(model)
    if not info.can_cache:
        pytest.skip(f"Model {model} can not be cached.")

    if model in MODELS_ADD_SKIP:
        pytest.skip(f"Model {model} is marked as un-testable.")

    if model == 'posm':
        pytest.skip(f"Model {model} has no shape parameter.")

    mfunc = getattr(x, model)

    n = 5
    pars = [np.full(n, 0.1) if p.name.casefold() == 'redshift'
            else np.full(n, p.default)
            for p in info.parameters]
    pars = np.column_stack(pars)

    egrid = np.arange(0.1, 10, 0.01)
    y1 = np.array([mfunc(energies=egrid, pars=p) for p in pars])
    y2 = jax.vmap(p, in_axes=(0, None, None), out_axes=0)(pars, egrid, 1)
    y3 = jax.jit(jax.vmap(p, in_axes=(1, None, None), out_axes=0))(pars.T, egrid, 1)

    assert np.allclose(y1, y2)
    assert np.allclose(y1, y3)


@pytest.mark.parametrize("model", MODELS_CON)
def test_batching_con(model):
    """Evaluate a convolution additive model."""
    p, info = x.get_primitive(model)
    if not info.can_cache:
        pytest.skip(f"Model {model} can not be cached.")

    if model in MODELS_CON_SKIP:
        pytest.skip(f"Model {model} is marked as un-testable.")

    egrid = np.arange(0.1, 10, 0.01)

    def conv(p):
        """what is the default parameter value?"""

        if p.name.casefold() == 'redshift':
            return 0.01

        if p.name.casefold() == 'velocity':
            return 100.0

        return p.default

    pars = [p.default for p in x.info('powerlaw').parameters]
    pars = np.array(pars)
    mvals = x.powerlaw(energies=egrid, pars=pars)
    # mvals = mvals.astype(get_dtype(info))
    ymodel = mvals.copy()

    mfunc = getattr(x, model)

    n = 5
    pars = np.column_stack([np.full(n, conv(p)) for p in info.parameters])
    y1 = np.array([mfunc(energies=egrid, pars=p, model=mvals) for p in pars])
    y2 = jax.vmap(p, in_axes=(0, None, None, None), out_axes=0)(pars, egrid, mvals, 1)
    y3 = jax.jit(jax.vmap(p, in_axes=(0, None, None, None), out_axes=0))(pars, egrid, mvals, 1)
    mvals_n = np.tile(mvals, (n, 1))
    y4 = jax.vmap(p, in_axes=(None, None, 0, None), out_axes=0)(pars[0], egrid, mvals_n, 1)
    y5 = jax.vmap(p, in_axes=(0, None, 0, None), out_axes=0)(pars, egrid, mvals_n, 1)

    assert (y1 > 0).any()
    assert np.any(y1 != ymodel)
    assert np.allclose(y1, y2)
    assert np.allclose(y1, y3)
    assert np.allclose(y1, y4)
    assert np.allclose(y1, y5)


# @pytest.mark.parametrize("model", MODELS_ADD + MODELS_MUL)
# def test_jvp(model):
#     p, info = x.get_primitive(model)
#     if not info.can_cache:
#         pytest.skip(f"Model {model} can not be cached.")
#
#     if model in MODELS_ADD_SKIP:
#         pytest.skip(f"Model {model} is marked as un-testable.")
#
#     if model == 'posm':
#         pytest.skip(f"Model {model} has no shape parameter.")
#
#     if model in ['nsa', 'nsmax', 'nsmaxg', 'nsx']:
#         pytest.skip()
#
#     pars = np.array([0.1 if p.name.casefold() == 'redshift' else p.default
#                      for p in info.parameters])
#
#     egrid = np.arange(0.1, 10, 0.01)
#
#     prim_in = (pars, egrid, 1)
#     tan_in = (
#         np.ones_like(pars),
#         np.ones_like(egrid),
#         np.zeros((), dtype=jax.dtypes.float0)
#     )
#
#     tan = jax.jvp(p, prim_in, tan_in)[1]
#
#     assert jnp.isfinite(tan).all()
#     assert jnp.isnan(tan).sum() == 0



import glob
import os
import pathlib
import platform
import sys
import sysconfig

from setuptools import setup

from pybind11.setup_helpers import Pybind11Extension, build_ext

sys.path.append(os.path.dirname(__file__))
from helpers import template
from helpers.identify_xspec import get_xspec_macros

__version__ = '0.0.4'

# Check HEASARC is set up. The following does not provide a useful
# message from 'pip install' so how do we make it more meaningful?
#
HEADAS = os.getenv('HEADAS')
if HEADAS is None:
    sys.stderr.write('ERROR: unable to find HEADAS environment variable.\n')
    sys.exit(1)

HEADAS = pathlib.Path(HEADAS)

modelfile = HEADAS / '../spectral/manager/model.dat'
modelfile = modelfile.resolve()

out_dir = pathlib.Path('src')
out_dir.mkdir(exist_ok=True)

# It would be nice to query for this from the system,
# such as with pkg_config. We can try and find the versions
# directly instead.
# xspec_libs = ['XSFunctions', 'XSUtil', 'XS', 'hdsp_6.29',
#               'cfitsio', 'CCfits_2.6', 'wcs-7.3.1']

# There's some attempt to be platform independent, but
# is it worth it?
libdir = HEADAS / sysconfig.get_config_var('platlibdir')
if platform.system() == 'Darwin':
    suffix = '.dylib'
else:
    suffix = sysconfig.get_config_var('SHLIB_SUFFIX')


# The tricky thing is that we have XSFunctions, XSUtil, and XS as
# arguments. So we cannot just look for XS*, as that will match
# multiple libraries. We also don't want to include all matches to XS
# as there are a number of matches we do not need.
def match(name):
    # Would it make sense to take the lib prefix from sysconfig?
    head = f'lib{name}{suffix}'
    ms = glob.glob(str(libdir / head))
    if len(ms) == 1:
        return name

    head = f'lib{name}_*{suffix}'
    ms = glob.glob(str(libdir / head))
    if len(ms) == 1:
        return pathlib.Path(ms[0]).stem[3:]

    head = f'lib{name}-*{suffix}'
    ms = glob.glob(str(libdir / head))
    if len(ms) == 1:
        return pathlib.Path(ms[0]).stem[3:]

    raise OSError(f'Unable to find a match for lib{name}*{suffix} in {libdir}')


xspec_libs = []
for libname in ['XSFunctions', 'XSUtil', 'XS', 'hdsp',
                'cfitsio', 'CCfits', 'wcs']:
    # Note: not all names are versioned
    xspec_libs.append(match(libname))


xspec_version, xspec_macros = get_xspec_macros(HEADAS)


# Create the code now we believe we have the Xspec installation
# sorted out.
info = template.apply(modelfile, xspec_version, out_dir)


# Note: we need access to the src/include directory - can we just
# hard-code this path or access it via some setuptools method?
include_dir = out_dir / 'include'
if not include_dir.is_dir():
    sys.stderr.write(f'ERROR: unable to find {include_dir}/')
    sys.exit(1)

macros = [('VERSION_INFO', __version__)] + xspec_macros


ext_modules = [
    Pybind11Extension(
        'xspex._compiled',
        [str(info['outfile'])],
        cxx_std=11,
        extra_compile_args=['-O3'],
        include_dirs=[str(include_dir), str(HEADAS / 'include')],
        library_dirs=[str(HEADAS / 'lib')],
        runtime_library_dirs=[str(HEADAS / 'lib')],
        libraries=xspec_libs,
        define_macros=macros
    )
]

setup(
    version=__version__,
    ext_modules=ext_modules,
    cmdclass={'build_ext': build_ext},
)
